import re
from simpleeval import simple_eval, NameNotDefined, AttributeDoesNotExist


class Expression:
    def __init__(self, expr):
        self.expr = expr

    def eval(self, context):
        raise NotImplementedError


class Bind(Expression):
    def eval(self, context):
        try:
            return simple_eval(self.expr, names=context)
        except AttributeDoesNotExist as e:
            obj = e.expression.split(e.attr)[0][:-1]  # page.valami; e.attr=valami
            if context.get(obj, None) is None:
                return None

        raise ValueError


class Condition(Bind):
    pass


class ForLoop(Expression):
    REGEX = re.compile(r'^\s*(.+)\s*in\s*(.+)\s*$')

    def __init__(self, expr):
        super().__init__(expr)
        m = self.REGEX.match(expr)
        if m:
            self.loop_variable = m.groups()[0].strip()
            if not self.loop_variable.isidentifier():
                raise SyntaxError(f'Not valid identifier name: {self.loop_variable}')
            self.iterable = m.groups()[1].strip()
        else:
            raise SyntaxError(f'For loop syntax error: {expr}')

    def eval(self, context):
        try:
            iterable = simple_eval(f'{self.iterable}', names=context)
            for counter, item in enumerate(iterable, start=1):
                yield counter, self.loop_variable, item
        except NameNotDefined as e:
            raise ValueError(f'"{self.iterable}" is not defined in context') from e



from django_brython.assets import require
from kirui.core import Component, Element, render

from browser import window, ajax, document, webcomponent


TEMPLATE = """
asdf sadfd
asdasdasd dddd eee fff ggg
"""
@Component(
    tag='kr-layout-sidebar',
    template=require('sidebar.html'),
    encapsulate_slot=False
)
class SidebarLayout(Element):
    _sidebar_is_open: bool = False
    _sidebar_is_moving: bool = False
    is_pwa: bool
    sidebar: "document.Element"
    main: "document.Element"

    def connectedCallback(self):
        print('connected23456789asdlkaasdassssdsdssdsdddssddssdsd')
        super().connectedCallback()



from browser import ajax, webcomponent, window, document


class TemplateLike:
    def connectedCallback(self):
        parent_config = getattr(self.parent, '_component_config', None)
        if parent_config:
            encapsulate_slot = getattr(parent_config, 'encapsulate_slot', None)
            if encapsulate_slot is None:
                encapsulate_slot = self.parent.attrs.get('data-encapsulate-slot', False)

            if encapsulate_slot:
                self.attachShadow({'mode': 'open'})


# webcomponent.define('kr-template', TemplateLike)



from browser import document, window


class VDOMRenderer:
    def render(self, velement: window.HTMLElement):
        tag_name = 'DIV' if velement.tagName == 'TEMPLATE' else velement.tagName
        element = document.createElement(tag_name)
        for vchild in velement.childNodes:
            if vchild.nodeType == 3:
                child = document.createTextNode(vchild.textContent)
            else:
                child = self.render(vchild)

            element.appendChild(child)

        return element

    def __call__(self, velement: window.HTMLElement):
        return self.render(velement)


render = VDOMRenderer()



import hashlib


def get_md5(fname):
    hash_md5 = hashlib.md5()
    try:
        with open(fname, "rb") as rfile:
            for chunk in iter(lambda: rfile.read(4096), b""):
                hash_md5.update(chunk)

            return hash_md5.hexdigest()
    except FileNotFoundError:
        pass



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..environment import Environment


class DictParser:
    def __init__(self, environment: 'Environment'):
        self.environment = environment

        self.template = None
        self.act_parent = None

    def parse(self, source: dict, template_name: str):
        self.template = self.environment.template_class()
        self.template.root_element = self.parseNode(**source)

        return self.template

    def parseNode(self, tagName, attrs, children):
        klass = self.environment.registry.get_class_by_name(tagName)

        namespaced_attrs = {}
        for k, v in attrs.items():
            namespaced_attrs[(None, k)] = v

        element = klass(xml_tag=tagName, xml_attrs=namespaced_attrs)

        for i, child in enumerate(children):
            if isinstance(child, str):
                # first and last empty text childs (i == 0 or i == len(children) - 1) and
                if len(child.strip()) == 0:
                    continue

                child = child.lstrip()

                parts = child.rsplit('\n', 1)
                if len(parts) == 2 and len(parts[1].strip()) == 0:
                    child = parts[0]

                child = self.environment.registry.anonymus_element_klass(text=child)
            else:
                child['tagName'] = child['tagName'].lower()
                child = self.parseNode(**child)

            element.add_child(child)

        return element



"""Runserver command with livereload"""
import re
from socket import error as SocketError
from multiprocessing import Process, Event
from django.conf import settings
from django.apps import apps
from django.core.management.base import CommandError
from django.core.management.commands.runserver import naiveip_re
from django.utils.module_loading import import_string

from livesync.asyncserver import LiveSyncSocketServer, dispatcher
from livesync.fswatcher import FileWatcher


if 'django.contrib.staticfiles' in settings.INSTALLED_APPS:
    from django.contrib.staticfiles.management.commands.runserver import \
        Command as RunserverCommand
else:
    from django.core.management.commands.runserver import Command as RunserverCommand


class Command(RunserverCommand):
    """
    Command for running the development server with LiveSyncReload.
    """
    help = 'Starts a lightweight Web server for development with Live Sync.'

    def __init__(self):
        super(Command, self).__init__()
        self.server = None
        self.server_process = None
        self.file_watcher = None
        self.init_settings()

    def init_settings(self):
        if not hasattr(settings, 'DJANGO_LIVESYNC'):
            settings.DJANGO_LIVESYNC = dict()

        settings.DJANGO_LIVESYNC.setdefault('PORT', 9001)
        settings.DJANGO_LIVESYNC.setdefault('HOST', 'localhost')
        settings.DJANGO_LIVESYNC.setdefault('EXCLUDED_APPS', {'admin', 'debug_toolbar', 'livesync', })
        settings.DJANGO_LIVESYNC.setdefault('INCLUDED_APPS', set(apps.app_configs.keys()) - settings.DJANGO_LIVESYNC['EXCLUDED_APPS'])
        settings.DJANGO_LIVESYNC.setdefault('EVENT_HANDLER', 'livesync.core.handler.LiveReloadRequestHandler')

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('--liveport', action='store', dest='liveport', default=9001,
                            help='Host and port on which Django LiveSync shoud run.'),

    @property
    def liveport(self):
        return settings.DJANGO_LIVESYNC['PORT']

    @liveport.setter
    def liveport(self, value):
        settings.DJANGO_LIVESYNC['PORT'] = value

    @property
    def livehost(self):
        return settings.DJANGO_LIVESYNC['HOST']

    @livehost.setter
    def livehost(self, value):
        settings.DJANGO_LIVESYNC['HOST'] = value

    def _start_async_server(self):
        self.server = LiveSyncSocketServer(port=self.liveport)
        server_started_event = Event()
        self.server_process = Process(target=self.server.start, args=(server_started_event,))
        self.server_process.daemon = True
        self.server_process.start()
        server_started_event.wait(timeout=0.1)
        return server_started_event.is_set()

    def _stop_async_server(self):
        if self.server_process:
            self.server.server_close()
            self.server_process.join()

    def _start_watchdog(self):
        klass = import_string(settings.DJANGO_LIVESYNC['EVENT_HANDLER'])
        self.file_watcher = FileWatcher()
        self.file_watcher.add_handler(klass())
        self.file_watcher.start()

    def _stop_watchdog(self):
        if self.file_watcher:
            self.file_watcher.stop()

    def _parse_options(self, **options):
        if options['liveport']:
            liveport = options['liveport']
            if not isinstance(liveport, int) and not liveport.isdigit():
                raise CommandError("LiveSync port must be an integer.")
            self.liveport = int(liveport)

        if options['addrport']:
            matches = re.match(naiveip_re, options['addrport'])
            if matches and matches.groups()[0]:
                self.livehost = matches.groups()[0]

    def stop_liveserver(self):
        self._stop_async_server()
        self._stop_watchdog()

    def start_liveserver(self, **options):
        self._parse_options(**options)

        if self._start_async_server():
            self._start_watchdog()

    def handle(self, *args, **options):
        self.start_liveserver(**options)
        super(Command, self).handle(*args, **options)
        self.stop_liveserver()



import json
from threading import Timer

from django.dispatch import receiver
from websocket import create_connection
from django.conf import settings

from livesync.core.signals import livesync_event

EVENTS = {
    'refresh': {"action": "refresh"}
}


@receiver(livesync_event)
def dispatch(event, **kwargs):
    uri = "ws://{host}:{port}".format(
        host=settings.DJANGO_LIVESYNC['HOST'],
        port=settings.DJANGO_LIVESYNC['PORT'])

    connection = create_connection(uri)
    if hasattr(event, '__iter__'):
        event = json.dumps(dict(event))
    elif isinstance(event, str):
        event = json.dumps(EVENTS[event])
    else:
        raise NotImplementedError

    connection.send(event)
    connection.close()



import sys
from io import StringIO

from .elements import BaseElement


class Template:
    def __init__(self):
        self.root_element = None  # type: BaseElement
        self.source_path = None  # type: Path

    def _show_element_subtree(self, element: BaseElement, stdout, indent: int=1):
        spaces = ' ' * (indent - 1) * 4
        print(f'{spaces} {element.__class__.__name__} <{element.xml_tag}>', file=stdout)
        for child in element.children:
            self._show_element_subtree(child, stdout=stdout, indent=indent + 1)

    def show_element_tree(self, stdout=sys.stdout):
        return self._show_element_subtree(element=self.root_element, stdout=stdout)

    def _as_xml_subtree(self, element: BaseElement, io, newline: str, indent: int=1):
        spaces = ' ' * (indent - 1) * 4

        if not element.xml_tag:  # anonymus elements
            print(f'{spaces}{element.text}', file=io, end=newline)
        else:
            closing = '>' if len(element.children) else '/>'
            print(f'{spaces}<{element.xml_tag}{element.attrs_as_xml}{closing}', file=io, end=newline)
            for i, child in enumerate(element.children):
                self._as_xml_subtree(child, io, newline=newline, indent=indent + 1)

            if len(element.children):
                print(f'{spaces}</{element.xml_tag}>', file=io, end=newline)

    def serialize(self, output='xml'):
        if output == 'xml':
            io = StringIO()
            self.root_element.xml_attrs = {
                'xmlns:c': "https://doculabs.io/2020/xtmpl#control",
                'xmlns:b': "https://doculabs.io/2020/xtmpl#data-binding"
            }
            self._as_xml_subtree(element=self.root_element, io=io, newline='\n', indent=1)
            return io.getvalue()
        else:
            raise NotImplementedError(f'Invalid output type: {output}')



from .core import render, patch_dom_with_vdom, minify

from browser import document


class App:
    el: 'document.HTMLElement'
    components: []

    def __init__(self, components):
        self.components = components

    def mount(self, selector):
        el = document.querySelector(selector)
        #el.style['display'] = 'block'
        #return

        vdom = document.createElement('DIV')
        for attr in el.attributes:
            vdom.setAttribute(attr.name, attr.value)

        minify(el.content)
        patch_dom_with_vdom(vdom, el.content.getRootNode())
        # print(vdom.innerHTML)
        # el.replaceWith(vdom)
        #document.body.appendChild(vdom)
        el.replaceWith(vdom)



from django.http import HttpRequest
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from django.template.backends.base import BaseEngine
from django.template.backends.utils import csrf_input_lazy, csrf_token_lazy
from django.template.context import make_context
from django.utils.functional import cached_property
from django.utils.module_loading import import_string

from django_kirui.components import KrApp, KrForm
from django_kirui.components.forms import KrField
from samon.environment import Environment
from samon.loaders import FileSystemLoader
from samon.exceptions import TemplateNotFound, TemplateError
from samon.template import Template


class DjangoSamonTemplates(BaseEngine):
    app_dirname = 'django_react_ssr'

    def __init__(self, params):
        params = params.copy()
        options = params.pop("OPTIONS", {}).copy()
        super().__init__(params)

        self.engine = Environment(loader=FileSystemLoader(search_path=self.template_dirs))
        self.engine.registry.element('kr-app', KrApp)
        self.engine.registry.element('kr-form', KrForm)
        self.engine.registry.element('kr-form-field', KrField)
        self._context_processors = options.pop('context_processors', [])

    @cached_property
    def context_processors(self):
        return tuple(import_string(path) for path in self._context_processors)

    def get_template(self, template_name):
        try:
            return Template(self.engine.get_template(template_name), backend=self)
        except TemplateNotFound as exc:
            raise TemplateDoesNotExist(exc.args, backend=self)
        except TemplateError as exc:
            raise TemplateSyntaxError(exc.args)


class Template:
    DEFAULT_OUTPUT_FORMAT = 'json'

    def __init__(self, template: Template, backend: DjangoSamonTemplates):
        self.template = template
        self.backend = backend
        # TODO: self.origin = ... # https://docs.djangoproject.com/en/3.1/ref/templates/api/#django.template.base.Origin

    def render(self, context: dict = None, request: HttpRequest = None):
        context = make_context(context, request, autoescape=True)

        output = self.DEFAULT_OUTPUT_FORMAT
        if request is not None:
            context['request'] = request
            context['csrf_input'] = csrf_input_lazy(request)
            context['csrf_token'] = csrf_token_lazy(request)

            if request.content_type in ('html/text', 'text/plain', ''):
                output = 'xml'
            else:
                output = 'json'

            # Support for django context processors
            for processor in self.backend.context_processors:
                context.update(processor(request))

        # return json.dumps(self.template.root_element.render(context).serialize(output=output))
        return self.template.root_element.render(context).serialize(output=output)
